From b83355ffd6babee0a71c5c892b1872a7601643e3 Mon Sep 17 00:00:00 2001
From: Wang Xingchao <xingchao.wang@linux.intel.com>
Date: Wed, 24 Jul 2013 17:38:33 +0800
Subject: [PATCH RFC] ALSA: hda - Change power-save policy for power-well
 regression

This patch intended to fix power-well regression on Haswell.

On Harris Beach(Ultrabook with battery), there's only eDP panel connected by default, no HDMI/DP.
And gfx driver needs enable some HW feature for eDP, power-well *must* be
disabled in this scenario.
- Witout charger connected, power-well feature is perfect
- with charger connected, audio never release power-well.

Basically, power-well should be release if audio driver doesnot use it, that's
why we enabled runtime power-save feature.

In second case above, with charger connected, the parameter under
"/sys/devices/../power/control" become "on" always.
In audio driver side, power_save was set to "0", which disable power_down the
codecs and controller, thus never release power->usage_count.

And this blocks audio driver release power-well.

In the second case, if audio driver detect hdmi pins are free and no Apps
opened device, it will eanble runtime power-save feature as an exception.

I test this patch on Harris beach with charger connected, the power-well could
be released as expected.

Signed-off-by: Wang Xingchao <xingchao.wang@linux.intel.com>
---
 sound/pci/hda/hda_codec.h  |    1 +
 sound/pci/hda/hda_intel.c  |   56 ++++++++++++++++++++++++++++++++++++++++++++
 sound/pci/hda/patch_hdmi.c |   31 ++++++++++++++++++++++++
 3 files changed, 88 insertions(+)

diff --git a/sound/pci/hda/hda_codec.h b/sound/pci/hda/hda_codec.h
index 701c2e0..7a2030d 100644
--- a/sound/pci/hda/hda_codec.h
+++ b/sound/pci/hda/hda_codec.h
@@ -726,6 +726,7 @@ struct hda_codec_ops {
 	int (*check_power_status)(struct hda_codec *codec, hda_nid_t nid);
 #endif
 	void (*reboot_notify)(struct hda_codec *codec);
+	bool (*codec_busy)(struct hda_codec *codec);
 };
 
 /* record for amp information cache */
diff --git a/sound/pci/hda/hda_intel.c b/sound/pci/hda/hda_intel.c
index 8860dd5..65afe11 100644
--- a/sound/pci/hda/hda_intel.c
+++ b/sound/pci/hda/hda_intel.c
@@ -544,6 +544,7 @@ struct azx {
 
 #ifdef CONFIG_SND_HDA_I915
 	struct work_struct probe_work;
+	struct delayed_work powerwell_work;
 #endif
 
 	/* reboot notifier (for mysterious hangup problem at power-down) */
@@ -739,6 +740,7 @@ static inline void mark_runtime_wc(struct azx *chip, struct azx_dev *azx_dev,
 
 static int azx_acquire_irq(struct azx *chip, int do_disconnect);
 static int azx_send_cmd(struct hda_bus *bus, unsigned int val);
+static int azx_power_save_update(struct azx *chip);
 /*
  * Interface for HD codec
  */
@@ -1978,6 +1980,7 @@ static int azx_pcm_open(struct snd_pcm_substream *substream)
 	struct azx_pcm *apcm = snd_pcm_substream_chip(substream);
 	struct hda_pcm_stream *hinfo = apcm->hinfo[substream->stream];
 	struct azx *chip = apcm->chip;
+	struct pci_dev *pci = chip->pci;
 	struct azx_dev *azx_dev;
 	struct snd_pcm_runtime *runtime = substream->runtime;
 	unsigned long flags;
@@ -2064,9 +2067,12 @@ static int azx_pcm_close(struct snd_pcm_substream *substream)
 	struct azx_pcm *apcm = snd_pcm_substream_chip(substream);
 	struct hda_pcm_stream *hinfo = apcm->hinfo[substream->stream];
 	struct azx *chip = apcm->chip;
+	struct pci_dev *pci = chip->pci;
 	struct azx_dev *azx_dev = get_azx_dev(substream);
 	unsigned long flags;
 
+	if (chip->driver_caps & AZX_DCAPS_I915_POWERWELL)
+		azx_power_save_update(chip);
 	mutex_lock(&chip->open_mutex);
 	spin_lock_irqsave(&chip->reg_lock, flags);
 	azx_dev->substream = NULL;
@@ -3444,6 +3450,52 @@ static void azx_probe_work(struct work_struct *work)
 {
 	azx_probe_continue(container_of(work, struct azx, probe_work));
 }
+
+static bool check_azx_free(struct azx *chip)
+{
+	struct azx_dev *azx_dev;
+	int i;
+
+	for (i = 0; i < chip->num_streams; i++) {
+		azx_dev = &chip->azx_dev[i];
+		if (azx_dev->prepared)
+			return true;
+	}
+	return false;
+}
+
+static int azx_power_save_update(struct azx *chip)
+{
+	struct pci_dev *pci = chip->pci;
+	struct hda_codec *codec;
+	int usage_count;
+	bool codec_busy; /* Check if pins connecting Monitors */
+
+	list_for_each_entry(codec, &chip->bus->codec_list, list) {
+		if (codec->patch_ops.codec_busy)
+			codec_busy = codec->patch_ops.codec_busy(codec);
+	}
+
+	usage_count = atomic_read(&pci->dev.power.usage_count);
+	if (usage_count > 0 &&
+			!check_azx_free(chip) &&
+			!codec_busy) {
+		pm_runtime_allow(&pci->dev);
+		power_save = 1;
+	}
+
+	return usage_count;
+}
+
+static void azx_powerwell_work(struct work_struct *work)
+{
+	struct azx *chip = container_of(work, struct azx, powerwell_work.work);
+	int repeat;
+
+	repeat = azx_power_save_update(chip);
+	if (repeat > 0)
+		schedule_delayed_work(&chip->powerwell_work, msecs_to_jiffies(5000));
+}
 #endif
 
 /*
@@ -3524,6 +3576,7 @@ static int azx_create(struct snd_card *card, struct pci_dev *pci,
 #ifdef CONFIG_SND_HDA_I915
 	/* continue probing in work context as may trigger request module */
 	INIT_WORK(&chip->probe_work, azx_probe_work);
+	INIT_DELAYED_WORK(&chip->powerwell_work, azx_powerwell_work);
 #endif
 
 	*rchip = chip;
@@ -3890,6 +3943,9 @@ static int azx_probe_continue(struct azx *chip)
 	if (chip->driver_caps & AZX_DCAPS_PM_RUNTIME)
 		pm_runtime_put_noidle(&pci->dev);
 
+	if (chip->driver_caps & AZX_DCAPS_I915_POWERWELL)
+		schedule_delayed_work(&chip->powerwell_work, msecs_to_jiffies(5000));
+
 	return 0;
 
 out_free:
diff --git a/sound/pci/hda/patch_hdmi.c b/sound/pci/hda/patch_hdmi.c
index 030ca86..a692078 100644
--- a/sound/pci/hda/patch_hdmi.c
+++ b/sound/pci/hda/patch_hdmi.c
@@ -1882,6 +1882,22 @@ static int generic_hdmi_resume(struct hda_codec *codec)
 }
 #endif
 
+bool generic_hdmi_busy(struct hda_codec *codec)
+{
+	struct hdmi_spec *spec = codec->spec;
+	struct hdmi_eld *eld;
+	int pin_idx;
+
+	for (pin_idx = 0; pin_idx < spec->num_pins; pin_idx++) {
+		eld = &get_pin(spec, pin_idx)->sink_eld;
+		if (eld->monitor_present ||
+				eld->eld_valid)
+			return true;
+	}
+	snd_printdd("HDMI pins free now\n");
+	return false;
+}
+
 static const struct hda_codec_ops generic_hdmi_patch_ops = {
 	.init			= generic_hdmi_init,
 	.free			= generic_hdmi_free,
@@ -1891,8 +1907,23 @@ static const struct hda_codec_ops generic_hdmi_patch_ops = {
 #ifdef CONFIG_PM
 	.resume			= generic_hdmi_resume,
 #endif
+	.codec_busy		= generic_hdmi_busy,
 };
 
+bool intel_haswell_pins_busy(struct hda_codec *codec)
+{
+	struct hdmi_spec *spec = codec->spec;
+	struct hdmi_eld *eld;
+	int pin_idx;
+
+	for (pin_idx = 0; pin_idx < spec->num_pins; pin_idx++) {
+		eld = &get_pin(spec, pin_idx)->sink_eld;
+		if (eld->monitor_present ||
+				eld->eld_valid)
+			return true;
+	}
+	return false;
+}
 
 static void intel_haswell_fixup_connect_list(struct hda_codec *codec,
 					     hda_nid_t nid)
-- 
1.7.9.5

